/* FUSH - $version$
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

   FUSH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   FUSH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/
#include "fush.h"
#include "linklist.h"
#include "md5.h"

extern List g_alias;
extern List g_conf;
extern List g_seconf;

/* Sat Nov 29 10:56:24 PST 2003
 * get a list of tokens, and expand variables and aliases */
int fu_expand( char* tokens, int num_tokens ) {
  int tpos;
  int opos;
  int temptok;
  int foundtok = num_tokens;
  int setme = 0;
  char *ptr;
  char *expanded;
  char *argument;

  if ( ( expanded = malloc( MAX_TOKENS ) ) == NULL )
    perror( "malloc()" );

  for ( tpos = 0; num_tokens >= 0; num_tokens-- ) {
    /* get the next argument */
    argument = &tokens[ tpos ];
    tpos += strlen( &tokens[ tpos ] ) + 1;

    /* this is an alias command */
    if ( setme == 0 ) {
      setme = 1;
      if ( ( ptr = find_alias( argument, &g_alias ) ) != NULL ) {
        temptok = fu_tokenize( ptr, expanded, 0 );
        foundtok += temptok;
        /* get the length of the expanded alias */
        for ( opos = 0; temptok > 0; temptok-- )
          opos += strlen( &expanded[ opos ] )+1;
        /* move enerything on the line, back */
        memmove( &tokens[ opos ], &tokens[ tpos ], MAX_TOKENS - opos );
        /* replace the token with the alias tokens */
        memcpy( argument, expanded, opos );
      }
    }

    if ( strcmp( argument, "~" ) == 0 ) {
      ptr = getenv( "HOME" ); 
      opos = strlen( ptr );
      /* move enerything on the line, back */
      memmove( &tokens[ tpos + opos ], &tokens[ tpos ], MAX_TOKENS - tpos );
      /* replace the token with the alias tokens */
      memcpy( argument, ptr, opos );
    }

    /* variable expansion here */
  }
  return ( foundtok );
}


/* Tue Nov 11 14:07:37 PST 2003
 * search though the string, take out tokens (white space, shell ops, etc)
 * and create a string of tokens where *tokens seperated by \0 */
int fu_tokenize( char *string, char *tokens, int allowcomment ) {

  char check;
  char *buffer = NULL;
  char *ptr = NULL;
  int fullength;
  int tpos = 0;
  int spos = -1;
  int bpos = 0;
  int ltok = 0;
  int numtok = 0;
  int single_quote = 0;
  int escape = 0;

  fullength = strlen( string );

  if ( fullength >= MAX_CMD )
    die( "can't parse strings that long" );

  if ( ( buffer = malloc( MAX_TOKENS ) ) == NULL )
    perror( "malloc()" );

  bzero( tokens, MAX_TOKENS );
  bzero( buffer, MAX_TOKENS );

  while ( spos++ < fullength ) {

    /* there should be no way that this condition is reached, but...
     * don't write past the end of the buffer and segfault */
    if ( tpos >= MAX_TOKENS - 10 )
      return ( -1 );

    /* current character */
    check = buffer[ bpos ] = string[ spos ];

    /* filter out format characters to syslog! */
    if ( check == '%' )
      buffer[ ++bpos ] = '%';

    bpos++;

    if ( check == 0 )
      goto fintok;

    /* XXX BROKEN */
    /* if escaped, then decrement escaped by one */
    if ( escape > 0 )
      escape--;

    /* turn on escaping */
    if ( check == '\\' && !escape && !single_quote ) {
      escape = 2;
      goto fintok;
    }

    /* if this char is escaped, then just copy it */
    if ( escape == 1 ) {
      tokens[ tpos ] = string[ spos ];
      tpos++;
      ltok = 0;
      goto fintok;
    }

    /* single quote, override everything else! */
    if ( single_quote ) {
      /* end a single quote */
      if ( check == '\'' || check == '\"' )
        single_quote = 0;
      else {
        tokens[ tpos ] = string[ spos ];
        tpos++;

        if ( check == '\\' ) {
          tokens[ tpos ] = string[ spos ];
          tpos++;
        }

        ltok = 0;
      }

      goto fintok;
    }

    /* if we should process comments, and we have a #, then fin process */
    if ( allowcomment && check == '#' )
      break;

    if ( check == '~' ) {
      ptr = getenv( "HOME" ); 
      strncpy( &tokens[ tpos ], ptr, MAX_TOKENS - tpos );
      tpos += strlen( ptr );
      ltok = 0;
      goto fintok;
    }  

    /* accounting for shell operators */
    if ( check == '|' || check == '<' || check == '>' || check == ';' || check == '&' ) {
      if ( !ltok ) {
        tokens[ tpos ] = '\0';
        tpos++;
        numtok++;
      }

      tpos += 2;
      numtok++;
      ltok = 1;
    }

    /* white space denotes tokens */
    if ( check == ' ' || check == '\t' ) {
      if ( !ltok ) {
        tokens[ tpos ] = '\0';
        ltok = 1;
        tpos++;
        numtok++;
      }
    }

    /* shell output operators */
    else if ( check == '>' ) {
      if ( ! fu_get_conf( &g_seconf, "fu_redirect_out", 0 ) ) {
        syndenied( "output redirection" );
        return ( -1 );
      }

      if ( string[ spos + 1 ] == '>' ) {
        memcpy( &tokens[ tpos - 2 ], ">>\0", 3 );
        tpos++;
        spos++;
      }
      else {
        memcpy( &tokens[ tpos - 2 ], ">\0", 2 );
      }
    }

    /* shell input operator */
    else if ( check == '<' ) {
      if ( ! fu_get_conf( &g_seconf, "fu_redirect_in", 0 ) ) {
        syndenied( "input redirection" );
        return ( -1 );
      }

      memcpy( &tokens[ tpos - 2 ], "<\0", 2 );
    }

    /* shell input operator */
    else if ( check == '|' ) {
      if ( string[ spos + 1 ] == '|' ) {
        if ( ! fu_get_conf( &g_seconf, "fu_logical_or", 0 ) ) {
          syndenied( "logical or" );
          return ( -1 );
        }
        memcpy( &tokens[ tpos - 2 ], "||\0", 3 );
        tpos++;
        spos++;
      }
      else {
        if ( ! fu_get_conf( &g_seconf, "fu_redirect_pipe", 0 ) ) {
          syndenied( "pipes" );
          return ( -1 );
        }

        memcpy( &tokens[ tpos - 2 ], "|\0", 2 );
      }
    }

    /* shell background operator */
    else if ( check == '&' ) {
      if ( string[ spos + 1 ] == '&' ) {
        if ( ! fu_get_conf( &g_seconf, "fu_logical_and", 0 ) ) {
          syndenied( "logical AND" );
          return ( -1 );
        }
        memcpy( &tokens[ tpos - 2 ], "&&\0", 3 );
        tpos++;
        spos++;
      }
      else {
        memcpy( &tokens[ tpos - 2 ], "&\0", 2 );
      }
    }

    /* shell ; operator */
    else if ( check == ';' ) {
      if ( ! fu_get_conf( &g_seconf, "fu_multi_command", 0 ) ) {
        syndenied( "multiple commands" );
        return ( -1 );
      }

      memcpy( &tokens[ tpos - 2 ], ";\0", 2 );
    }

    /* start a single quote */
    else if ( ( check == '\'' || check == '\"' ) && ! single_quote )
      single_quote = 1;

    /* kill the end of the line on comments */
    else if ( check == '#' && ! single_quote )
      break;

    /* dump the character to the tokenized string */
    else {
      tokens[ tpos ] = string[ spos ];
      tpos++;
      ltok = 0;
    }

fintok:
    if ( spos == fullength )
      numtok++;
  }

  if ( fu_get_conf( &g_seconf, "fu_syslog", 0 ) )
    syslog( LOG_INFO, buffer );
  free( buffer );

  /* don't count white space as a token */
  if ( ltok ) {
    numtok--;
  }

  return ( numtok );
}


/* Fri Nov 14 18:54:15 PST 2003
 * glob an expression out into arguments */
int fu_glob( glob_t *globby, char *expression ) {

  int num_glob;
  int conf;

  if ( globby->gl_pathc )
    num_glob = glob( expression, GLOB_APPEND | GLOB_DOOFFS | GLOB_ERR | GLOB_NOCHECK | GLOB_NOSORT, NULL, globby );
  else
    num_glob = glob( expression, GLOB_DOOFFS | GLOB_ERR | GLOB_NOCHECK | GLOB_NOSORT, NULL, globby );

  if ( num_glob < 0 )
    perror( "GLOB()" );

  conf = atoi( fu_get_conf( &g_seconf, "fu_max_args", 0 ) );
  if ( globby->gl_pathc > conf ) {
    syndenied( "maximum number of arguments exceeded" );
    globfree( globby );
    bzero( globby, sizeof( globby ) );
    return ( -1 );
  }

  return ( num_glob );
}


/* Tue Nov 11 00:14:57 PST 2003
 * foneyop
 * parse the option, command, and return argument type */
int get_state( char* argument ) {

  /* internal command - cd */
  if ( strcmp( argument, "cd" ) == 0 )
    return ( S_CD );

  /* internal command - cd */
  if ( strcmp( argument, "chdir" ) == 0 )
    return ( S_CD );

  /* internal command - pwd */
  else if ( strcmp( argument, "pwd" ) == 0 )
    return ( S_PWD );

  /* internal command - umask */
  else if ( strcmp( argument, "umask" ) == 0 )
    return ( S_UMASK );

  /* internal command - alias */
  else if ( strcmp( argument, "alias" ) == 0 )
    return ( S_ALIAS );

  /* internal command - alias */
  else if ( strcmp( argument, "unalias" ) == 0 )
    return ( S_UNALIAS );

  /* internal command - setenv */
  else if ( strcmp( argument, "putenv" ) == 0 )
    return ( S_EXPORT );
  else if ( strcmp( argument, "setenv" ) == 0 )
    return ( S_EXPORT );
  else if ( strcmp( argument, "export" ) == 0 )
    return ( S_EXPORT );
  else if ( strcmp( argument, "set" ) == 0 )
    return ( S_SET );
  else if ( strpbrk( argument, "=" ) != NULL )
    return ( S_SETVALUE );

  /* ineternal command - unset */
  else if ( strcmp( argument, "unset" ) == 0 )
    return ( S_UNSET );
  else if ( strcmp( argument, "unsetenv" ) == 0 )
    return ( S_UNENV );

  /* exit commands */
  else if ( strcasecmp( argument, "exit" ) == 0 )
    return ( S_QUIT );
  else if ( strcasecmp( argument, "logout" ) == 0 )
    return ( S_QUIT );

  /* internal operator - redirect overwrite */
  else if ( strcmp( argument, ">" ) == 0 )
    return ( S_RDROVR );
  else if ( strcmp( argument, "<" ) == 0 )
    return ( S_RDRIN );
  else if ( strcmp( argument, ">>" ) == 0 )
    return ( S_RDRAPD );
  else if ( strcmp( argument, "|" ) == 0 )
    return ( S_PIPE );
  else if ( strcmp( argument, ";" ) == 0 )
    return ( S_LASTOPT );
  else if ( strcmp( argument, "&" ) == 0 )
    return ( S_NOWAIT );
  else if ( strcmp( argument, "&&" ) == 0 )
    return ( S_AND );
  else if ( strcmp( argument, "||" ) == 0 )
    return ( S_OR );
  else if ( argument[ 0 ] == '\0' )
    return ( S_LASTOPT );

  return ( S_UNKNOWN );
}


/* Sun Nov  9 12:08:33 PST 2003
 * */
char *get_prompt( char* prompt, int plen ) {
  char buffer[ MAXPATHLEN ];
  char cwd[ MAXPATHLEN ];
  char buffer2[ MAXPATHLEN ];
  char sec[ 3 ];
  char min[ 3 ];
  char hour[ 3 ];
  char mday[ 3 ];
  char mon[ 3 ];
  char year[ 5 ];
  char *username;
  char *hostname;
  char *fullname;
  char *prompt_env;
  char *conf;
  int exec_ctr = 0;
  int conlen = 0;
  int freeme = 0;
  uid_t id;
  time_t rawtime;

  struct tm *timeinfo;


  bzero( buffer, MAXPATHLEN );
  bzero( cwd, MAXPATHLEN );
  bzero( prompt, plen );

  /* get the time */
  time ( &rawtime );
  timeinfo = localtime ( &rawtime );

  /* get the hostname */

  if ( gethostname( buffer, BUFFMEDM ) == -1 )
    die( "can't get hostname" );

  fullname = strdup( buffer );

  hostname = buffer;
  if ( !( hostname = strsep( &hostname, "." ) ) )
    hostname = fullname;

  /* get the username */
  if ( !( username = getlogin( ) ) )
    die( "getlogin()" );

  /* get the cwd */

  id = getuid();

  /* convert time from int to ascii */
  snprintf( sec, 3, "%d", timeinfo->tm_sec );

  snprintf( min, 3, "%d", timeinfo->tm_min );

  snprintf( mon, 3, "%d", timeinfo->tm_mon + 1 );

  snprintf( hour, 3, "%d", timeinfo->tm_hour );

  snprintf( mday, 3, "%d", timeinfo->tm_mday );

  snprintf( year, 5, "%d", timeinfo->tm_year + 1900 );

  if ( ( prompt_env = fu_get_conf( &g_conf,  "prompt", 0 ) ) == NULL ) {
    prompt_env = malloc( BUFFMEDM );
    bzero( prompt_env, BUFFMEDM );
    sprintf( prompt_env, "%%m%%# " );
    freeme = 1;
  }

  while ( exec_ctr < strlen( prompt_env ) ) {
    if ( prompt_env[ exec_ctr ] == '%' ) {
      exec_ctr++;

      /* %M - FQDN */
      if ( prompt_env[ exec_ctr ] == 'M' ) {
        strncpy( &prompt[ conlen ], fullname, plen - conlen );
        conlen += strlen( fullname );
      } 
      
      /* %m - hostname */
      else if ( prompt_env[ exec_ctr ] == 'm' ) {
        strncpy( &prompt[ conlen ], hostname, plen - conlen );
        conlen += strlen( hostname );
      } 
      
      /* %n - username */
      else if ( prompt_env[ exec_ctr ] == 'n' ) {
        strncpy( &prompt[ conlen ], username, plen - conlen );
        conlen += strlen( username );
      } /* %% - a % sign */
      else if ( prompt_env[ exec_ctr ] == '%' ) {
        prompt[ conlen ] = prompt_env[ exec_ctr ];
        conlen++;
      } /* %# - a > for users a # for root */
      else if ( prompt_env[ exec_ctr ] == '#' ) {
        if ( id == 0 )
          prompt[ conlen ] = '#';
        else
          prompt[ conlen ] = '>';

        conlen++;
      } /* %? - the return code of the last exec command */
      else if ( prompt_env[ exec_ctr ] == '?' ) {
        conf = fu_get_conf( &g_conf, "status", 0 );
        strncpy( &prompt[ conlen ], conf, plen - conlen );
        conlen += strlen( conf );
      } /* %/ - print the current working directory */
      else if ( prompt_env[ exec_ctr ] == '/' ) {
        strncpy( &prompt[ conlen ], cwd, plen - conlen );
        conlen += strlen( cwd );
      } 

      /* %~ - print the current working directory */
      else if ( prompt_env[ exec_ctr ] == '~' ) {
        if ( !( getcwd( buffer, MAXPATHLEN ) ) )
          perror( "getcwd()" );
        conf = fu_get_conf( &g_conf, "home", 0 );

        /* replace home dir with ~ */
        if( strncmp( buffer, conf, strlen( conf ) ) == 0 ) {
          strcpy ( buffer2, "~" );
          strncpy( &buffer2[ 1 ], &buffer[ strlen( conf ) ], 
                   MAXPATHLEN - 1 );
        }
        else
          strncpy( buffer2, buffer, MAXPATHLEN );

        strncpy( &prompt[ conlen ], buffer2, plen - conlen );
        conlen += strlen( buffer2 );
      }

       /* %L - print a newline character */
      else if ( prompt_env[ exec_ctr ] == 'L' ) {
        prompt[ conlen ] = '\n';
        conlen++;
      } /* %s - the number of seconds since last minute */
      else if ( prompt_env[ exec_ctr ] == 's' ) {
        strncpy( &prompt[ conlen ], sec, plen - conlen );
        conlen += strlen( sec );
      } /* %i - the minute since last hour  */
      else if ( prompt_env[ exec_ctr ] == 'i' ) {
        strncpy( &prompt[ conlen ], min, plen - conlen );
        conlen += strlen( min );
      } /* %h - the hour elapsed this day  */
      else if ( prompt_env[ exec_ctr ] == 'h' ) {
        strncpy( &prompt[ conlen ], hour, plen - conlen );
        conlen += strlen( hour );
      } /* %w - day of month */
      else if ( prompt_env[ exec_ctr ] == 'w' ) {
        strncpy( &prompt[ conlen ], mon, plen - conlen );
        conlen += strlen( mon );
      } /* %D - print a newline character */
      else if ( prompt_env[ exec_ctr ] == 'D' ) {
        strncpy( &prompt[ conlen ], mday, plen - conlen );
        conlen += strlen( mday );
      } /* %Y - print a newline character */
      else if ( prompt_env[ exec_ctr ] == 'Y' ) {
        strncpy( &prompt[ conlen ], year, plen - conlen );
        conlen += strlen( year );
      }
    }
    else {
      prompt[ conlen ] = prompt_env[ exec_ctr ];
      conlen++;
    }

    exec_ctr++;
  }
  if( freeme ) {
    free( prompt_env );
  }

  return ( prompt );
}



/* Sat Nov 15 17:20:36 PST 2003
 * read a line from the fd passed to us */
int fu_readline( int fd, char *line, int len ) {

  char *ptr = line;
  int num_read = 0;

  bzero( line, len );

  /* bail on obvious error */
  if ( !fd )
    return ( -1 );

  /* read until we find a \n */
  while ( read( fd, ptr, 1 ) == 1 ) {
    num_read++;
    
    if ( *ptr == '\n' )
      break;
    ptr++;
  }
  *ptr = '\0';

  /* return EOF on EOF */
  if ( *ptr == EOF ) {
    return ( EOF );
  }

  /* return number of bytes read */
  return ( num_read );
}


