/* FUSH - $version$ 
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

    FUSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FUSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/

#ifndef FUSH_H

#ifdef HAVE_CONFIH_H
#include "../config.h"
#endif

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <glob.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef TM_IN_SYS_TIME_H
#include <sys/time.h>
#else 
#include <time.h>
#endif

#include "linklist.h"

#define S_CD       1
#define S_PWD      2
#define S_EXPORT   3
#define S_UNSET    4
#define S_RDROVR   5
#define S_RDRAPD   6
#define S_RDRIN    7
#define S_PIPE     8
#define S_GLOB     9
#define S_NOWAIT   10
#define S_UMASK    11
#define S_AND      12
#define S_OR       13
#define S_SET      14
#define S_ALIAS    15
#define S_UNALIAS  16
#define S_UNENV    17
#define S_SETVALUE 18
#define S_INTERR   240
#define S_QUIT     241
#define S_WAITNEXT 252
#define S_LASTCMD  253
#define S_LASTOPT  254
#define S_NOP      255
#define S_UNKNOWN  256


#define F_NOHUP    1
#define F_BACK     2
#define F_PIPE     4
#define F_NOEXEC   8
#define F_HAVEBOR  16
#define F_HAVEBAND 32
#define F_ALIAS    64
#define F_NOKILL   128


#define MAX_CMD    	4096 
#define MAX_TOKENS 	8192 

#ifndef MAXPATHLEN
#define MAXPATHLEN  1024
#endif

#define CMD_SIZE 50 

#define BUFFTINY 32
#define BUFFSMAL 64
#define BUFFMEDM 256
#define BUFFLARG 512
#define BUFFXLRG 1024
#define BUFFMAXM 4096

/* Mon Nov 10 20:39:31 PST 2003
 * This is a backgrounded process name and id number */
typedef struct back_cmd_ {
  char  command[MAXPATHLEN];
  int   c_num; 
  pid_t pid;
} back_cmd;

/* Mon Dec  1 20:08:51 PST 2003
 * an alias command XXX cahnge to value_pair*/
typedef struct alias_cmd_ {
  char name[MAXPATHLEN];
	char command[MAX_TOKENS]; 
	int  len;
} alias_cmd;

/* Tue Nov 18 08:51:18 MST 2003
 * This is a command in the username.security and fush.db files */
typedef struct list_cmd_ {
  unsigned char md5[35];
	int  length;
  char command;
} list_cmd;

/* Mon Dec  1 20:04:09 PST 2003
 * a generic name value pair */
typedef struct value_pair_ {
	char name[BUFFSMAL];
  char value[BUFFLARG];
} value_pair;



void die( char * );
void fu_exit( int );
void syndenied( char * );
void process_cmds( void );
int fuwrite( int, void *, int );
int get_state( char* );
int fu_tokenize( char *, char *, int );
int md5file( char* , char *);
char *get_prompt( char*, int );
back_cmd* find_proc( List *, int );
int fu_glob( glob_t *, char * );
int read_security( char* );
void proc_resource( char* );
int fu_readline( int fd, char *line, int len );
ListElmt* find_cmd( char *, char *, List*);
void read_db( char* );
int process_tokens( char *, int );
char* find_alias( char *, List* );
int fu_unalias( char* );
int can_exec( List *, char *);
int fu_expand( char *, int );
void fu_show_alias( List* );
void fu_do_opts( int, char ** );
int fu_add_value( List *, char *, char * );
void fu_conf_default( List * );
char* fu_get_conf( List *, char *, int );
int find_path( char *, char *, int );
char* resolve_path( char *command, char *realpath, int length );
void fu_clean_env( List *cleanenv );
int fu_dup_tty( void );

#endif
