/* FUSH - $VERSION$
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

   FUSH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   FUSH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "fush.h"
#include <signal.h>
#include <pwd.h>
#include <dirent.h>
#include <ctype.h>
#include <regex.h>
#include "md5.h"


#ifdef HAVE_LIBREADLINE
#include <readline/history.h>
#include <readline/readline.h>
#endif

regex_t r_shells;
regex_t r_perl;
regex_t r_tcl;
regex_t r_python;
regex_t r_emacs;

/* Sat Nov  8 14:24:53 PST 2003
 * Foneyop
 * read the command in from stdin */
char * get_command( char* command, char *prompt ) {
#ifndef HAVE_LIBREADLINE
  int pos = 0;
#endif

  if ( command != NULL ) {
    free( command );
    command = NULL;
  }


#ifndef HAVE_LIBREADLINE

  if ( ( command = malloc( MAX_CMD ) ) == NULL )
    die( "no momory" );

  bzero( command, MAX_CMD );

  write( 1, prompt, strlen( prompt ) );

  while ( pos < MAX_CMD ) {
    read( 0, &command[ pos ], 1 );

    if ( command[ pos ] == 10 || command[ pos ] == '\0' )
      break;

    pos++;
  }

  command[ pos-- ] = '\0';
#endif

#ifdef HAVE_LIBREADLINE

  command = readline( prompt );

  if ( command && *command )
    add_history( command );

#endif

  return ( command );
}



/* Sat Nov  8 22:42:04 PST 2003
 * die with error */
void die( char *message ) {
  perror( message );
  exit( -1 );
}


/* Wed Nov 19 10:57:30 MST 2003
 *   */
int is_restricted( char *file ) {

  int error = 0;
  char *tok = NULL;
  char *last = NULL;
  char *brk = NULL;
  char copy[ MAXPATHLEN ];

  error = strlen( file );
  file[ error - 1 ] = '\0';

  error = regexec( &r_shells, file, 0, NULL, 0 );

  if ( !error ) {
    printf( "not adding command: [%s] - looks like a shell\n", file );
    return ( 1 );
  }

  error = regexec( &r_perl, file, 0, NULL, 0 );

  if ( !error ) {
    printf( "not adding command: [%s] - looks like a perl interpreter\n", file );
    return ( 1 );
  }

  error = regexec( &r_tcl, file, 0, NULL, 0 );

  if ( !error ) {
    printf( "not adding command: [%s] - looks like a tcl interpreter\n", file );
    return ( 1 );
  }

  error = regexec( &r_python, file, 0, NULL, 0 );

  if ( !error ) {
    printf( "not adding command: [%s] - looks like a python interpreter\n", file );
    return ( 1 );
  }

  error = regexec( &r_emacs, file, 0, NULL, 0 );

  if ( !error ) {
    printf( "not adding command: [%s] - looks like an emacs editor\n", file );
    return ( 1 );
  }

  strncpy( copy, file, MAXPATHLEN );

  tok = strtok_r( copy, "/", &brk );

  while ( ( tok = strtok_r( NULL, "/", &brk ) ) != NULL ) {
    last = tok;
  }

  if ( strcmp( last, "find" ) == 0 ) {
    printf( "not adding command: [%s] - looks like find command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "xargs" ) == 0 ) {
    printf( "not adding command: [%s] - looks like xargs command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "expect" ) == 0 ) {
    printf( "not adding command: [%s] - looks like expect command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "awk" ) == 0 ) {
    printf( "not adding command: [%s] - looks like awk command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "nohup" ) == 0 ) {
    printf( "not adding command: [%s] - looks like nohup command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "wish" ) == 0 ) {
    printf( "not adding command: [%s] - looks like wish command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "nice" ) == 0 ) {
    printf( "not adding command: [%s] - looks like nice command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "time" ) == 0 ) {
    printf( "not adding command: [%s] - looks like time command\n", file );
    return ( 1 );
  }

  if ( strcmp( last, "." ) == 0 ) {
    return ( 1 );
  }
  if ( strcmp( last, ".." ) == 0 ) {
    return ( 1 );
  }

  return ( 0 );
}





/* Fri Nov 14 16:47:56 MST 2003
 * Converted digest to 32byte ascii string (less computation on load) */
int md5file( char *digest, char *file ) {
  int fd;
  int numread;
  int i;
  char *errorstring = NULL;
  char *buffer = NULL;
  unsigned char binary[ 17 ];
  struct MD5Context context;

  bzero( binary, 17 );
  bzero( digest, 35 );

  MD5Init( &context );

  if ( ( buffer = malloc( 8192 ) ) == NULL ) {
    errorstring = strerror( errno );
    syslog( LOG_ERR, errorstring );
    return ( -1 );
  }

  if ( ( fd = open( file, O_RDONLY ) ) == -1 )
    return ( -1 );

  while ( ( numread = read( fd, buffer, 8192 ) ) > 0 )
    MD5Update( &context, buffer, numread );

  MD5Final( binary, &context );
  free( buffer );
  if( close( fd ) == -1 )
    perror( "close()" );

  for ( i = 0; i <= 16; i++ ) {
    sprintf( &digest[ i * 2 ], "%02x", binary[ i ] );
  }

  return ( 0 );
}


/* Wed Nov 19 10:56:12 MST 2003
 * calculate a list of MD5 sums for all binaries */
int calc_md5s( char *path, int outfd, int out2fd ) {
  DIR * dirp;

  struct dirent *dp;
  char digest[ 35 ];
  char buffer[ 1024 ];
  int slen;

  bzero( buffer, 1024 );
  bzero( digest, 35 );

  if ( ( dirp = opendir( path ) ) == NULL ) {
    perror( "opendir()" );
    return ( 0 );
  }

  while ( ( dp = readdir( dirp ) ) != NULL ) {

    bzero( buffer, 1024 );
    bzero( digest, 17 );
    snprintf( buffer, 1024, "%s/%s", path, dp->d_name );

    if ( md5file( digest, buffer ) != 0 ) {
      printf( "md5 error: [%s]\n", buffer );
      return ( -1 );
    }

    slen = strlen( buffer );
    sprintf( &buffer[ slen ], "\n" );
    write( outfd, buffer, slen + 1 );

    if ( !is_restricted( buffer ) ) {
      sprintf( &buffer[ slen ], "\n" );
      write( out2fd, buffer, slen + 1 );
    }

    bzero( buffer, 256 );

    digest[ 33 ] = '\n';
    write( outfd, digest, 34 );
  }

  closedir( dirp );
  return( 0 );
}


/* Sat Nov  8 22:43:53 PST 2003
 * set stuff up and call main loop
 */
int main( int argc, char **argv ) {
  uid_t id;
  int outfd;
  int out2fd;
  char *ptr = NULL;
  char *ptrbuffer = NULL;
  char *tok = NULL;
  char *line = NULL;
  char buffer[ 8194 ];
  char path[ 1024 ];
  char *username;
  struct passwd *passwd = NULL;

  bzero( buffer, 8194 );
  //bzero( digest, 17 );

  /* open a connection to syslog */
  if ( !( username = getlogin( ) ) )
    die( "getlogin()" );

  openlog( username, LOG_NDELAY | LOG_PID, LOG_LOCAL0 );

  snprintf( buffer, 255, "user: %s started fushadmin", username );

  syslog( LOG_NOTICE, buffer );

  id = getuid();

  //if ( id != 0 )
   // die( "you must be root to run fushadmin" );

  /* exec a command if arg -c */
  if ( argv[ 1 ] != NULL ) {
    if ( strcmp( argv[ 1 ], "-c" ) == 0 ) {
      syslog( LOG_NOTICE, "fush admin reconfigure" );

      printf( "Configuring fush...\n" );
      ptr = getenv( "PATH" );
      ptrbuffer = strdup( ptr );
      tok = strtok( ptrbuffer, ":" );

      snprintf( buffer, BUFFMEDM, "Where to write output? enter='%s'> ", _CONFIGDIR_ );
      line = get_command( line, buffer );
      if ( strlen( line ) < 1 )
        sprintf( path, "%s", _CONFIGDIR_ );
      else
        snprintf( path, 256, "%s", line );
      umask( 0022 );
      mkdir( path, 0777 );

      snprintf( buffer, BUFFMEDM, "%s/fush.db", path );
      if ( ( outfd = open( buffer, O_WRONLY | O_CREAT, 0666 ) )
           < 0 )
        die( "can't open fush.db" );

      snprintf( buffer, BUFFMEDM, "%s/fush.security", path );
      if ( ( out2fd = open( buffer, O_WRONLY | O_CREAT, 0666 ) )
           < 0 )
        die( "can't open fush.security" );

      regcomp( &r_shells, ".*sh.?[\\.0-9]*$", REG_EXTENDED );
      regcomp( &r_perl, ".*perl.*", 0 );
      regcomp( &r_tcl, ".*tcl.*", 0 );
      regcomp( &r_python, ".*python.*", 0 );
      regcomp( &r_emacs, ".*emac.*", 0 );

      snprintf( buffer, 8192, "# This is a fush .security template file\n\
# This file was created by fuadmin\n\
##########################################################################\n\
\n\
# allow command chains\n\
fu_redirect_pipe=no\n\
\n\
# allow output redirection\n\
fu_redirect_in=no\n\
\n\
# allow output redirection\n\
fu_redirect_out=no\n\
\n\
# allow multiple commands on the command line\n\
fu_multi_command=no\n\
\n\
# restrict the logical OR function\n\
fu_logical_or=yes\n\
\n\
# restrict the logical AND function\n\
fu_logical_and=yes\n\
\n\
# allow user to change directory\n\
fu_chdir=yes\n\
\n\
# allow user to execute full paths\n\
fu_exec_full_path=yes\n\
\n\
# allow user to login interactivly to the system\n\
fu_allow_interactive=yes\n\
\n\
# set to the main system wide profile (cshrc, zshenv, bashrc, etc...)\n\
# special note: fush will not correctly process if then statements etc...\n\
# (set to 'no' to disable)\n\
fu_system_rc=/etc/profile\n\
\n\
# home system profile (the users home dir is auto prepended)\n\
# (set to 'no' to disable)\n\
fu_home_rc=.profile\n\
\n\
# enable home .fushrc profile (yes or no)\n\
fu_use_homerc=yes\n\
\n\
# enable system _CONFIGDIR_/fushrc profile (yes or no)\n\
fu_use_systemrc=yes\n\
\n\
# maximum number of backgrounded commands\n\
fu_max_background=1\n\
\n\
# nice value for backgrounded commands\n\
fu_background_nice=15\n\
\n\
# set ACL to whitelist\n\
# NOTE: not setting a value on limit_exec turns off command checks\n\
fu_limit_exec=include\n\
\n\
# Allow user to exec commands owned by user\n\
fu_exec_owner=no\n\
\n\
# maximum arguments on the command line\n\
fu_max_args=512\n\
\n\
# log stdin to syslog\n\
fu_syslog=yes\n\
# facility to log to LOCAL0 - LOCAL9\n\
fu_log_facility=128\n\
# log stdout to file (filename - fullpath)\n\
fu_log_output=no\n\
\n\
\n\
# command substitutions, by command name\n\
subst_command vi=rvim\n\
subst_command vim=rvim\n\
subst_command gvim=rvim\n\
# redirect emacs users to vim (hehe)\n\
subst_command emacs=rvim\n\
\n\
# command substitutions, by fullpath\n\
subst_path /bin/sh=/usr/local/bin/fush\n\
\n\
\n\
# policy enforced environment variables\n\
set_env LD_RUN_PATH=\n\
set_env DT_RUN_PATH=\n\
set_env DT_RPATH=\n\
set_env LD_LIBRARY_PATH=\n\
set_env SHELL=/usr/local/bin/fush\n\
set_env PAGER=more\n\
\n\
\n\
# Turn on owner restriction, set to no to turn off\n\
fu_restrict_owner=yes\n\
# Only files owned by these users can be executed\n\
allow_owners=root bin\n\
# Only files owned by these groups can be executed\n\
allow_groups=root wheel bin\n\
\n\
\n\
# command ACL\n\
# commands must be listed as full paths\n\
##########################################################################\n" );

    write( out2fd, buffer, strlen( buffer ) );

    while ( tok != NULL ) {
      snprintf( buffer, BUFFMEDM, "Allow commands in dir: [%s] [yes|no]? ", tok );
      line = ( char * ) get_command( line, buffer );
      if ( strncasecmp( line, "y", 1 ) == 0 || line[0] == 0 ) {
        calc_md5s( tok, outfd, out2fd );
      }
      tok = strtok( NULL, ":" );
    }

    close( outfd );
    close( out2fd );
    return( 0 );
  }
    if ( strcmp( argv[ 1 ], "-u" ) == 0 ) {
      printf("configuring users!\n");

      snprintf( buffer, BUFFMEDM, "Where should I write output? enter='%s'> ", _CONFIGDIR_ );
      line = get_command( line, buffer );
      if ( strlen( line ) < 1 )
        sprintf( path, "%s", _CONFIGDIR_ );
      else
        snprintf( path, 256, "%s", line );

      while ( ( passwd = getpwent( ) ) != NULL ) {
        printf("configuring user [%s]\n", passwd->pw_name);
        snprintf( buffer, BUFFMEDM, "cp %s/fush.security %s/%s.security", path, path, passwd->pw_name );
        system( buffer );
      }
    }
    return( 0 );
  }

  return ( -1 );
}
