/* FUSH - $VERSION$
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

   FUSH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   FUSH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/
#include "../config.h"
#include "fush.h"
#include "linklist.h"
#include "md5.h"
#include <pwd.h>
#include <libutil.h>
#include <termios.h>

extern List g_replace;
extern List g_alias;
extern List g_conf;
extern List g_seconf;
extern List g_procs;
extern List g_exec;
extern List g_db;
extern List g_setenv;
extern List g_owner;
extern List g_group;

/* Fri Nov 14 16:47:56 MST 2003
 * Converted digest to 32byte ascii string (less computation on load) */
int md5file( char *digest, char *file ) {
  int fd;
  int numread;
  int i;
  char *errorstring = NULL;
  char *buffer = NULL;
  unsigned char binary[ 17 ];
  struct MD5Context context;

  bzero( binary, 17 );
  bzero( digest, 35 );

  if ( ( fd = open( file, O_RDONLY ) ) == -1 ) {
    return ( -1 );
  }

  MD5Init( &context );

  if ( ( buffer = malloc( 8192 ) ) == NULL ) {
    errorstring = strerror( errno );
    syslog( LOG_ERR, errorstring );
    return ( -1 );
  }

  while ( ( numread = read( fd, buffer, 8192 ) ) > 0 ) {
    MD5Update( &context, buffer, numread );
  }

  MD5Final( binary, &context );
  free( buffer );

  for ( i = 0; i <= 16; i++ ) {
    sprintf( &digest[ i * 2 ], "%02x", binary[ i ] );
  }

  return ( 0 );
}


/* Sat Nov  8 22:42:04 PST 2003
 * die with error */
void die( char *message ) {
  perror( message );
  fu_exit( -1 );
}

/* Fri Nov 14 09:13:48 MST 2003
 * free memory and exit */
void fu_exit( int status ) {
  printf("\n");
/* XXX BROKEN */
/*  list_destroy( &g_procs );
  list_destroy( &g_alias );
  list_destroy( &g_exec );
  list_destroy( &g_db ); */
  exit( status );
}

/* Wed Nov 12 20:25:35 PST 2003
 * tell the user the command was denied and call process_cmds( ) */
void syndenied( char *message ) {
  fprintf( stderr, "function [%s] violates system policy.\n", message );
  syslog( LOG_WARNING, "function [%s] violates system policy.", message );
  return ;
}


/* Sun Nov 30 14:40:18 MST 2003
 * remove an alias from the list of aliases */  
int fu_unalias( char *name ) {
  alias_cmd *aliascommand = NULL;
  void *data;
  ListElmt  *rain = list_head( &g_alias );
  ListElmt  *drop = NULL;

  while ( rain != NULL ) {
    aliascommand = rain->data;

    if ( strcmp( aliascommand->name, name ) == 0 ) {
      list_rem_next( &g_alias, drop, (void **)&data );
      free( data );
      return( 0 );
    }

    drop = rain;
    rain = rain->next;
  }

  return ( 1 );
}

/* Sat Nov 15??? I duno it was late...
 * check if the command is white listed or balcklisted */
int can_exec( List *list, char *command ) {
  char *path   = NULL;
  char *buffer = NULL;
  char *ptr    = NULL;
  char *conf   = NULL;
  char cwd[ MAXPATHLEN ];
  char search[ MAXPATHLEN ];
  char md5sum[ 33 ];
  ListElmt * value = NULL;
  struct stat sb;
  int user = 0;

  /* return true if we are configed not to check */
  conf = fu_get_conf( &g_seconf, "fu_limit_exec", 0 );
  if ( !conf )
    return ( 1 );

  bzero( md5sum, 33 );
  bzero( search, MAXPATHLEN + 1 );

  /* This is a full path */
  if ( command[ 0 ] == '/' || command[ 0 ] == '.' ) {
    strncpy( search, command, MAXPATHLEN );

    if ( md5file( md5sum, search ) != 0 )
      return ( 2 );
  }

  /* search the path */
  else {
    if ( ( path = getenv( "PATH" ) ) == NULL ) {
      getcwd(cwd, MAXPATHLEN );
      path = cwd;
    }

    buffer = strdup( path );

    if ( ( ptr = strsep( &buffer, ":" ) ) == NULL )
      die( "path error!" );

    while ( ( ptr = strsep( &buffer, ":" ) ) != NULL ) {
      snprintf( search, MAXPATHLEN, "%s/%s", ptr, command );

      if ( md5file( md5sum, search ) == 0 )
        break;
    }
    if( ptr == NULL )
      return( 2 );
  }

  /* get owner, get group */
  if ( fu_get_conf( &g_seconf, "fu_restrict_owner", 0 ) ) {
    stat( search, &sb );

    ptr = user_from_uid( sb.st_uid, 1 );
    if ( !fu_get_conf( &g_owner, ptr, 0 ) ) {
      user = getuid();  
      buffer = user_from_uid( user, 1 );
      if ( ( conf = fu_get_conf( &g_seconf, "fu_exec_owner", 0 ) ) &&
      strcmp( ptr, buffer ) != 0 )
        return ( 0 );
      else 
        user = -1;
    }
    ptr = user_from_uid( sb.st_gid, 1 );
    if ( !fu_get_conf( &g_group, ptr, 0 ) ) {
      if ( user != -1 )
        return ( 0 );
    }
  }

  value = find_cmd( search, md5sum, list );

  if ( value != NULL && ( strcasecmp(conf, "include") == 0 ) )
    return ( 1 );
  else if ( value == NULL && ( strcasecmp(conf, "exclude") == 0 ) )
    return ( 1 );
  else
    return ( 0 );
  return ( 2 );
}


/* 12/07/03 <time>
 * returns the full path of *command which exists in the environmet's PATH 
 * stores the full path in passed arg (fullpath) does not exceed length */
int find_path( char *command, char *realpath, int length ) {
  char *path     = NULL;
  char *ptr      = NULL;
  char *buffer   = NULL;
  char fullpath[MAXPATHLEN];
  int fd = 0;

  /* OBVIOUS! */
  if ( command == NULL )
    return 0;

  /* local path commands */
  if ( command[0] == '.' ) {
    path = &command[1];
    getcwd( fullpath, MAXPATHLEN );
    snprintf( realpath, length, "%s%s", fullpath, path );
    return( 1 );
  }

  /* full path commands */
  else if ( command[0] == '/' ) {
    snprintf( realpath, length, "%s", command );
    return( 1 );
  }

  /* get path, and make a copy */
  path = getenv( "PATH" );
  if ( path == NULL )
    return( 0 );
  buffer = strdup( path );

  /* search the path for the command */
  while ( ( ptr = strsep( &buffer, ":" ) ) != NULL ) {
    snprintf( fullpath, MAXPATHLEN, "%s/%s", ptr, command );
    if ( readlink( fullpath, realpath, length ) == -1 );
      strncpy( realpath, fullpath, length );
    if ( ( fd = open( realpath, O_NONBLOCK ) ) > 0 ) {
      close( fd );
      return( 1 );
    }
  }

  /* can't find this command in path! */
  return( 0 );
}


/* 12/07/03 <time>
 * returns the full path of *command which exists in the environmet's PATH 
 * NOTE: CALLER MUST FREE RETURNED POINTER */
char* resolve_path( char *command, char *realpath, int length ) {
  char *ptr      = NULL;
  char *file     = NULL;
  char *pptr     = NULL;
  char fullpath[MAXPATHLEN];
  value_pair *kv = NULL;
  ListElmt *rain = list_head( &g_replace );

  /* the returned command must be allocated with malloc, freed by freeglob */
  file = malloc( MAXPATHLEN );
  find_path( command, realpath, length );

  /* replace path names from our replace list */
  while ( rain != NULL ) {
    kv = rain->data;
    /* return the found key value */
    if( strcmp( kv->name, realpath ) == 0 ) {
      snprintf( realpath, length, "%s", kv->value );
      break;
    }
    rain = rain->next;
  }
  //printf("res [%s] from [%s]\n", realpath, command);

  snprintf( fullpath, MAXPATHLEN, "%s", realpath );
  /* get the absolute filename */
  pptr = fullpath;
  while ( ( ptr = strsep( &pptr, "/" ) ) != NULL )
    snprintf( file, MAXPATHLEN, "%s", ptr );

  /* this command was replaced with a fullpath, we are done */
  if ( rain != NULL )
    return( file );

  /* replace command names from our replace list */
  rain = list_head( &g_replace );
  while ( rain != NULL ) {
    kv = rain->data;
    /* return the found key value */
    if( strcmp( kv->name, file) == 0 ) {
      snprintf( file, MAXPATHLEN, "%s", kv->value );
      find_path( file, realpath, length );
      break;
    }
    rain = rain->next;
  }

  /* be sure to free(3) this! */
  return( file );
}





/* Fri Nov 28 16:29:21 PST 2003
 * display the list of aliases  */
void fu_show_alias( List* alias ) {
  alias_cmd * aliascommand = NULL;
  ListElmt *rain = list_head( alias );

  while ( rain != NULL ) {
    aliascommand = rain->data;
    printf( "%s\t(%s)\n", aliascommand->name, aliascommand->command );

    rain = rain->next;
  }

  return;
}

/* Sat Nov 15 19:16:32 PST 2003
 * return true if command is in our list of commands */
ListElmt* find_cmd( char *command, char *md5sum, List* searchlist ) {
  list_cmd *listcommand = NULL;
  ListElmt *rain = list_head( searchlist );

  while ( rain != NULL ) {
    listcommand = rain->data;

    if ( command != NULL && strcmp( command, &listcommand->command ) == 0 ) {
      if ( md5sum != NULL && bcmp( listcommand->md5, md5sum, 32 ) == 0 ) {
        return ( rain );
      } else if ( md5sum == NULL ) {
        return ( rain );
      }
    }

    rain = rain->next;
  }

  return ( NULL );
}

/* Wed Nov 12 21:38:58 PST 2003
 * */
back_cmd* find_proc( List *list, int value ) {
  int open = 0;
  back_cmd *cmd = NULL;
  back_cmd *rtr = NULL;
  back_cmd *prev = NULL;
  ListElmt *rain = list_head( list );
  ListElmt *plist = NULL;

  // Don't iterate over a crap list!
  if ( rain == NULL )
    return ( 0 );

  // loop over the list
  while ( rain != NULL ) {
    cmd = rain->data;

    // get the largest value
    if ( ! value ) {
      if ( cmd->c_num >= open ) {
        rtr = rain->data;
        open = cmd->c_num + 1;
      }
    }

    // get the matching pid
    else {
      if ( cmd->pid == ( int ) value ) {
        list_rem_next( list, plist, ( void* ) prev );
        return ( cmd );
      }
    }

    prev = cmd;
    plist = rain;
    rain = rain->next;
  }

  return ( rtr );
}


/* Fri Nov 28 19:40:06 PST 2003
 * return true if command is in our list of aliases */
char* find_alias( char *command, List* list ) {
  alias_cmd * alias = NULL;
  ListElmt *rain = list_head( list );

  if ( command == NULL )
    return ( NULL );

  while ( rain != NULL ) {
    alias = rain->data;

    if ( strcmp( command, alias->name ) == 0 ) {
      return ( alias->command );
    }

    rain = rain->next;
  }

  return ( NULL );
}


/* Tue Nov 18 17:40:57 MST 2003
 * process a security file 
 * Sat Nov 29 PST 2003 CoryMarsh: Changed memory alloc for list_cmd struct
 */
void read_db( char* file ) {
  list_cmd *listcommand = NULL;
  char *ptr = NULL; /* temp pointer */
  int infd = 0;     /* input FD */
  int z = 0;        /* temp integer */

  if ( ( infd = open( file, O_RDONLY ) ) == -1 )
    die( "can't open MD5 database" );

  if ( ( ptr = malloc( MAXPATHLEN ) ) == NULL )
    die( "malloc()" );

  /* continue reading lines from the input file until EOF */
  while ( fu_readline( infd, ptr, MAXPATHLEN ) > 0 ) {
    /* get the path of the command */
    if ( ptr[ 0 ] == '/' ) {
      z = CMD_SIZE + strlen( ptr );
      if ( ( listcommand = ( list_cmd * ) malloc( z ) ) == NULL )
        die( "malloc()" );


      bzero( listcommand, z );
      strncpy( &listcommand->command, ptr,  z - CMD_SIZE );

      listcommand->length = z;
    }
    /* get the MD5 hash of the command */
    else if ( listcommand != NULL ) {
      strncpy( listcommand->md5, ptr, 35 );
      list_ins_next( &g_db, list_tail( &g_exec ), listcommand );
      listcommand = NULL;
      z = 0;
    }
    //else
    //  fprintf( stderr, "The MD5 database is corrupt.  Please run fuadmin.\n" );
  }

  return ;
}


/* Tue Nov 25 11:46:57 MST 2003
 * process the users resource script */
void proc_resource( char* file ) {
  int infd = 0;
  int numtok = 0;
  char *ptr = NULL;
  char *tokens = NULL;

  if ( ( infd = open( file, O_RDONLY ) ) == -1 )
    return ;

  if ( ( ptr = malloc( MAXPATHLEN ) ) == NULL )
    die( "malloc()" );

  if ( ( tokens = malloc( MAX_TOKENS ) ) == NULL )
    perror( "malloc()" );

  while ( fu_readline( infd, ptr, MAXPATHLEN ) > 0 ) {
    numtok = fu_tokenize( ptr, tokens, 1 );
    process_tokens( tokens, numtok );
  }

  free( tokens );
  return ;
}



/* Sat Nov 15 14:44:04 PST 2003
 * process a security file */
int read_security( char* file ) {
  list_cmd *listcommand = NULL;
  list_cmd *cmdptr = NULL;
  ListElmt *element = NULL;
  char *ptr = NULL;
  char *bak = NULL;
  char *tok = NULL;
  char *val = NULL;
  int infd = 0;
  int z    = 0;

  if ( ( infd = open( file, O_RDONLY ) ) == -1 )
    return( -1 );

  if ( ( ptr = malloc( MAXPATHLEN ) ) == NULL )
    die( "malloc()" );
  bzero( ptr, MAXPATHLEN );

  while ( fu_readline( infd, ptr, MAXPATHLEN ) > 0 ) {
  
    bak = ptr;
    if( ( strncasecmp( bak, "subst_command", 13 ) == 0 ) ) {
      if( ( tok = strsep( &bak, " \t" ) ) && ( tok = strsep( &bak, "= \t" ) ) &&
      ( val = strsep( &bak, "= \t" ) ) ) {
        fu_add_value( &g_replace, tok, val );
      }
    }
    else if( ( strncasecmp( bak, "subst_path", 10 ) == 0 ) ) {
      if( ( tok = strsep( &bak, " \t" ) ) && ( tok = strsep( &bak, "= \t" ) ) &&
      ( val = strsep( &bak, "= \t" ) ) ) {
        fu_add_value( &g_replace, tok, val );
      }
    }
    else if( ( strncasecmp( bak, "set_env", 7 ) == 0 ) ) {
      if( ( tok = strsep( &bak, " \t" ) ) && ( tok = strsep( &bak, "= \t" ) ) &&
      ( val = strsep( &bak, "\n" ) ) ) {
        fu_add_value( &g_setenv, tok, val );
      }
    }
    else if( ( strncasecmp( bak, "allow_owners", 12 ) == 0 ) ) {
      if( ( tok = strsep( &bak, "=" ) ) != NULL ) {
        while ( ( val = strsep( &bak, " \t" ) ) != NULL )
          if( strlen(val) )
            fu_add_value( &g_owner, val, "blah" );
      }
    }
    else if( ( strncasecmp( bak, "allow_groups", 12 ) == 0 ) ) {
      if( ( tok = strsep( &bak, "=" ) ) != NULL ) {
        while ( ( val = strsep( &bak, " \t" ) ) != NULL )
          if( strlen(val) )
            fu_add_value( &g_group, val, "blah" );
      }
    }
    /* add a value pair to the seconf list */
    else if( ( tok = strsep( &bak, "=" ) ) && ( val = strsep( &bak, "=" ) ) )
      fu_add_value( &g_seconf, tok, val );

    /* add an entry to the exec list */
    else if ( ptr[ 0 ] == '/' ) {
      z = CMD_SIZE + strlen( ptr );
      if ( ( listcommand = ( list_cmd * ) malloc( z ) ) == NULL )
        die( "malloc()" );

      bzero( listcommand, z );
      strncpy( &listcommand->command, ptr, z - CMD_SIZE );

      if ( ( element = find_cmd( ptr, NULL, &g_db ) ) != NULL ) {
        cmdptr = element->data;
        memcpy( listcommand->md5, cmdptr->md5, 35 );

        list_ins_next( &g_exec, list_tail( &g_exec ), listcommand );
      }
      else if ( md5file( listcommand->md5, ptr ) == 0 ) {
        //fprintf( stderr, "Don't have secure MD5 for: %s\n", tok );
        list_ins_next( &g_exec, list_tail( &g_exec ), listcommand );
      }
      else
        free( listcommand );
    }
  }

  free( ptr );
  close( infd );
  return( 0 );
}


/* Mon Dec  1 19:10:17 PST 2003
 * process command line options */
void fu_do_opts( int argc, char **argv ) {
  char *cmds = NULL;
  char *tokens = NULL;
  int i = 0;
  int col = 0;
  int numtok;

  /* exec a command if arg -c */
  if ( argv[ 1 ] != NULL ) {
    if ( strcmp( argv[ 1 ], "-c" ) == 0 ) {
      cmds = ( char* ) malloc( MAX_CMD );
      bzero( cmds, MAX_CMD );
      if ( ( tokens = malloc( MAX_TOKENS ) ) == NULL )
        perror( "malloc()" );

      for ( i = 2, col = 0; i < argc; i++ ) {
        strncpy( &cmds[ col ], argv[ i ], MAX_CMD - col );
        col += strlen( argv[ i ] );
        cmds[ col++ ] = ' ';
      }

      numtok = fu_tokenize( cmds, tokens, 0 );
      process_tokens( tokens, numtok );
      free( tokens );
      exit( 0 );
    }
  }
}


/* Mon Dec  1 19:55:28 PST 2003
 * Setup configuration defaults */
void fu_conf_default( List *config ) {
  struct passwd *passwd = NULL;
  char buffer[BUFFLARG];
  char *ptr = NULL;
  int col = 0;

  /* default for output redirection */
  fu_add_value( &g_seconf, "fu_redirect_pipe", "yes" );

  /* default for output redirection */
  fu_add_value( &g_seconf, "fu_redirect_out", "yes" );

  /* default for input redirection */
  fu_add_value( &g_seconf, "fu_redirect_in", "yes" );

  /* default for multiple commands */
  fu_add_value( &g_seconf, "fu_multi_commnad", "yes" );

  /* default for multiple commands */
  fu_add_value( &g_seconf, "fu_use_rvim", "yes" );

  /* default for logical or  */
  fu_add_value( &g_seconf, "fu_logical_or", "yes" );

  /* default for logical and */
  fu_add_value( &g_seconf, "fu_logical_and", "yes" );

  /* default for system resource script */
  fu_add_value( &g_seconf, "fu_system_rc", "/etc/profile" );

  /* default for home resource script */
  fu_add_value( &g_seconf, "fu_home_rc", ".profile" );

  /* default for command restriction */
  fu_add_value( &g_seconf, "fu_limit_exec", "include" );

  /* default number of background commands */
  fu_add_value( &g_seconf, "fu_max_background", "1" );

  /* default number of background commands */
  fu_add_value( &g_seconf, "fu_background_nice", "15" );

  /* default number of maximum arguments */
  fu_add_value( &g_seconf, "fu_max_args", "256" );

  /* default for ability to change directory */
  fu_add_value( &g_seconf, "fu_chdir", "yes" );

  /* default for ability to execute full paths */
  fu_add_value( &g_seconf, "fu_exec_full_path", "yes" );

  /* default for always resolve full paths before execution */
  fu_add_value( &g_seconf, "fu_hard_path", "yes" );

  /* get the home directory */
  while ( ( passwd = getpwent( ) ) != NULL ) {
    if ( passwd->pw_uid == geteuid() ) {
      ptr = strdup( passwd->pw_dir );
      fu_add_value( config, "user", passwd->pw_name );

      /* get the home dir, follow symlinks and such */
      while( ( ptr = strsep( &passwd->pw_dir, "/" ) ) != NULL ) {
        buffer[col] = '/';
        col++;
        strncpy( &buffer[col], ptr, BUFFLARG - col);
        if( readlink( buffer, buffer, BUFFLARG - col ) == -1)
           strncpy( &buffer[col], ptr, BUFFLARG - col);
        col = strlen( buffer );
      }
      fu_add_value( config, "home", buffer );
      break;
    }
  }

  if( passwd == NULL )
    die( "you are not a valid user!" );

}


/* Mon Dec  1 20:51:30 PST 2003
 * Add a key value pair to a list */
int fu_add_value( List *conflist, char *key, char *value ) {

  value_pair *kv = NULL;

  if( ( kv = ( value_pair *) malloc( sizeof( value_pair ) ) ) == NULL ) {
    perror( "malloc(3)" );
    return( -1 );
  }
  
  bzero( kv, sizeof( value_pair ) );
  snprintf( kv->name,  BUFFMEDM, "%s", key );
  snprintf( kv->value, BUFFLARG, "%s", value );

  if( strlen( kv->name ) < 1 )
    return( -1 );

  /* delete this value */
  fu_get_conf( conflist, kv->name, 1 );

  /* if the value is no, don't reset it */
  if( strcasecmp( value, "no" ) != 0 )
    return( list_ins_next( conflist, list_tail( conflist ), kv ) );
  return( -1 );
}


/* Dec 3 <time> 
 * */
char* fu_get_conf( List *conflist, char *name, int delme ) {
  void *data;
  value_pair *kv = NULL;
  ListElmt *old  = list_head( conflist );
  ListElmt *rain = list_head( conflist );

  while ( rain != NULL ) {
    kv = rain->data;

    //printf("check name: [%s]\n", kv->name);
    /* display all results */
    if( name == NULL )
      printf( "%s\t%s\n", kv->name, kv->value );
    /* return the found key value */
    else if( strcmp( kv->name, name) == 0 && !delme )
      return( kv->value );
    /* remove the value */
    else if( strcmp( kv->name, name) == 0 && delme == 1 )
      list_rem_next( conflist, old, (void **)&data );

    old  = rain;
    rain = rain->next;
  }

  return ( NULL );
}

/* Thu Dec 18 19:25:31 PST 2003
 * force the environment on the user */
void fu_clean_env( List *cleanenv ) {
  char *name = NULL;
  value_pair *kv = NULL;
  ListElmt *rain = list_head( cleanenv );

  while ( rain != NULL ) {
    kv = rain->data;

    if( strlen( kv->value ) < 1 )
      unsetenv( name );
    else {
      unsetenv( kv->name );
      setenv( kv->name, kv->value, 1 );
    }

    rain = rain->next;
  }
}

/* Mon Dec 22 21:00:31 PST 2003
 * fork a process, seteuid and dup everything going to stdout to a file */
int fu_dup_tty( void ) {

  int  master = 0;
  int  slave = 0;
  int  numread = 0;
  int  fd = 0;
  int  in = 0;
	pid_t parrent = 0;
  pid_t child = 0;
  char *conf = NULL;
  char *ptr  = NULL;
  char buffer[ BUFFMEDM ];
  struct termios tt, rtt;
  struct winsize win;
  struct timeval *tvp = NULL;
  fd_set rfd;
  uid_t uid  = 0;
  uid_t euid = 0;


  if ( ( conf = fu_get_conf( &g_seconf, "fu_log_output", 0 ) ) == NULL )
    return( 0 );


  tcgetattr( STDIN_FILENO, &tt );
  ioctl( STDIN_FILENO, TIOCGWINSZ, &win );
  if ( openpty( &master, &slave, NULL, &tt, &win ) == -1 )
    die( "can't openpty(2)" );
  rtt = tt;
  cfmakeraw( &rtt );
  rtt.c_lflag &= ~ECHO;
  ( void )tcsetattr( STDIN_FILENO, TCSAFLUSH, &rtt );

	parrent = getpid();
  child = fork();
  if ( child < 0 )
    die( "fork()" );
  if ( child > 0 ) {
    
    /* regaid setuid privs */
    uid = getuid();
    ptr = fu_get_conf( &g_seconf, "euid", 0 );
    euid = atoi( ptr );
    errno = 0;
    if ( uid == euid )
      die( "fush is not setuid" );
    if ( ( in = seteuid( euid ) ) == -1 )
      die( "can't set effective userid!" );
    ptr = fu_get_conf( &g_seconf, "egid", 0 );
    euid = atoi( ptr );
    in = setegid( euid );

    umask( 0077 );
    if ( ( fd = open( conf, O_CREAT | O_WRONLY | O_APPEND, 0600 ) ) == -1 )
      die( "can't open(2) terminal output file" );
		snprintf( buffer, BUFFMEDM,
			"*** NEW FUSH SESSION:  PARRENT PROCESS: [%d] CHILD PROCESS: [%d] ***\n",
			 parrent, child );
		write( fd, buffer, strlen( buffer ) );
    FD_ZERO( &rfd );
    while ( 1 ) {
      FD_SET( master, &rfd );
      FD_SET( STDIN_FILENO, &rfd );

      in = select( master + 1, &rfd, 0, 0, tvp );
      if ( in < 0 && errno != EINTR )
        break;
      if ( in > 0 && FD_ISSET( STDIN_FILENO, &rfd ) ) {

        numread = read( STDIN_FILENO, buffer, BUFFMEDM );
        if ( numread <= 0 )
          break;
        if ( numread > 0 )
          write( master, buffer, numread );
      }

      if ( in > 0 && FD_ISSET( master, &rfd ) ) {
        numread = read( master, buffer, BUFFMEDM );
        if ( numread <= 0 )
          break;
        if ( numread > 0 ) {
          write( fd, buffer, numread );
          write( 1, buffer, numread );
        }
      }
    }
    exit( 0 );
  }

  if ( close( master ) == -1 )
    die( "can't close master terminal" );
  if ( login_tty( slave ) == -1 )
    die( "can't allocate new terminal" );

  return( 1 );
}

