/* FUSH - $VERSION$
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

   FUSH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   FUSH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/
#include "fush.h"
#include <stdlib.h>
#include <string.h>

#include "linklist.h"

/* setup a linkedlist */
/* FoneyOp 2/29/02 */
void list_init( List * list, void ( *destroy ) ( void *data ) ) {
  list->size = 0;
  list->destroy = destroy;
  list->head = NULL;
  list->tail = NULL;

  return ;
}


/* destroy a linked list */
/* FoneyOp 2/29/02 */
void list_destroy( List * list ) {
  void * data;

  /* don't pass control to NULL */
  if ( list == NULL )
    return ;

  /* while we have elements in the list */
  while ( list_size( list ) > 0 ) {
    if ( list_rem_next( list, NULL, ( void ** ) & data ) == 0 ) {
      if ( list->destroy != NULL )
        list->destroy( data );
    }
  }

  /* destroy list structure so it can not be used */
  memset( list, 0, sizeof( List ) );

  return ;
}


/* insert an emelent into the list */
/* FoneyOp 2/29/02 */
int list_ins_next( List * list, ListElmt * element, const void *data ) {
  ListElmt * new_element;

  /* allocate memory for list element */

  if ( ( new_element = ( ListElmt * ) malloc( sizeof( ListElmt ) ) ) == NULL )
    return ( -1 );

  /* set data in the list */
  new_element->data = ( void * ) data;

  /* instering element at head of list */
  if ( element == NULL ) {
    /* if this is the only element it is also the tail */

    if ( list_size( list ) == 0 )
      list->tail = new_element;

    new_element->next = list->head;

    list->head = new_element;
  }
  else {
    /* this is the last element in list, set to tail */
    /* note: if we lose part of the tail, that memory can not be free() */

    if ( element->next == NULL )
      list->tail = new_element;

    new_element->next = element->next;

    element->next = new_element;
  }

  /* add 1 to size of list */
  list->size++;

  return ( 0 );
}


/* remove an element from the list */
/* FoneyOp 2/29/02*/
int list_rem_next( List * list, ListElmt *element, void **data ) {

  ListElmt * old_element;

  /* bomb if we have no more elements in list */

  if ( list_size( list ) < 1 )
    return ( -1 );

  /* remove the head of the list */
  if ( element == NULL ) {
    data = list->head->data;
    old_element = list->head;
    list->head = list->head->next;

    /* if this is the tail, remove it */
    if ( list_size( list ) == 1 )
      list->tail = NULL;
  }
  else {
    /* if element we are removing does not have pointer then bomb */
    if ( element->next == NULL )
      return ( -1 );

    *data = element->next->data;
    old_element = element->next;
    element->next = element->next->next;

    if ( element->next == NULL )
      list->tail = element;
  }

  /* free the space for the list member structure */
  free( old_element );

  list->size--;

  return ( 0 );
}
